@@ -2,7 +2,7 @@ package Pcore::Chrome::Tab;
 
 use Pcore -class, -res;
 use Pcore::Lib::Data qw[to_json from_json from_b64];
-use Pcore::Lib::Scalar qw[weaken is_plain_coderef];
+use Pcore::Lib::Scalar qw[weaken];
 use Pcore::WebSocket::raw;
 
 use overload    #
@@ -23,8 +23,9 @@ has listen => ();    # {method => callback} hook
 has network_enabled => ( 0, init_arg => undef );
 has page_enabled    => ( 0, init_arg => undef );
 
-has _ws => ();       # websocket connection
-has _cb => ();       # { msgid => callback }
+has _ws  => ();                                              # websocket connection
+has _cb  => ();                                              # { msgid => callback }
+has _sem => sub { Coro::Semaphore->new(1) }, is => 'lazy';
 
 our $_MSG_ID = 0;
 
@@ -34,7 +35,7 @@ sub new_tab ( $self, @args ) {
     return;
 }
 
-sub close ( $self ) {    ## no critic qw[Subroutines::ProhibitBuiltinHomonyms NamingConventions::ProhibitAmbiguousNames]
+sub close ( $self ) {                                        ## no critic qw[Subroutines::ProhibitBuiltinHomonyms NamingConventions::ProhibitAmbiguousNames]
     return P->http->get("http://$self->{chrome}->{host}:$self->{chrome}->{port}/json/close/$self->{id}");
 }
 
@@ -42,25 +43,25 @@ sub activate ( $self ) {
     return P->http->get("http://$self->{chrome}->{host}:$self->{chrome}->{port}/json/activate/$self->{id}");
 }
 
-sub listen ( $self, $event, $cb = undef ) {    ## no critic qw[Subroutines::ProhibitBuiltinHomonyms]
+sub listen ( $self, $event, $cb = undef ) {                  ## no critic qw[Subroutines::ProhibitBuiltinHomonyms]
     if ($cb) {
         $self->{listen}->{$event} = $cb;
     }
     else {
-        delete $self->{listen}->{$event};
+        return delete $self->{listen}->{$event};
     }
 
     return;
 }
 
-sub _call ( $self, $method, $args = undef, $cb = undef ) {
+sub _call ( $self, $method, $args = undef ) {
     my $h = $self->{_ws} // $self->_connect;
 
     my $id = $_MSG_ID++;
 
     my $cv = P->cv;
 
-    $self->{_cb}->{$id} = sub ( $res ) { $cv->( $cb ? $cb->($res) : $res ) };
+    $self->{_cb}->{$id} = $cv;
 
     $self->{_ws}->send_text(
         \to_json {
@@ -70,13 +71,15 @@ sub _call ( $self, $method, $args = undef, $cb = undef ) {
         }
     );
 
-    return defined wantarray ? $cv->recv : ();
+    return $cv->recv;
 }
 
 sub _connect ( $self ) {
     weaken $self;
 
-    my $h = Pcore::WebSocket::raw->connect(
+    my $guard = $self->_sem->guard;
+
+    return $self->{_ws} ||= Pcore::WebSocket::raw->connect(
         "ws://$self->{chrome}->{host}:$self->{chrome}->{port}/devtools/page/$self->{id}",
         connect_timeout  => 1000,
         max_message_size => 0,
@@ -129,90 +132,64 @@ sub _connect ( $self ) {
             return;
         },
     );
-
-    $self->{_ws} = $h;
-
-    return $h;
 }
 
 # NETWORK
-sub network_enable ( $self, $cb = undef ) {
-    return $cb ? $cb->( res 200 ) : res 200 if $self->{network_enabled};
+sub network_enable ( $self ) {
+    return res 200 if $self->{network_enabled};
 
-    return $self->_call(
-        'Network.enable',
-        undef,
-        sub ($res) {
-            $self->{network_enabled} = 1 if $res;
+    my $res = $self->_call('Network.enable');
 
-            return $cb ? $cb->($res) : $res;
-        }
-    );
+    $self->{network_enabled} = 1 if $res;
+
+    return $res;
 }
 
-sub network_disable ( $self, $cb = undef ) {
-    return $cb ? $cb->( res 200 ) : res 200 if !$self->{network_enabled};
+sub network_disable ( $self ) {
+    return res 200 if !$self->{network_enabled};
 
-    return $self->_call(
-        'Network.disable',
-        undef,
-        sub ($res) {
-            $self->{network_enabled} = 0 if $res;
+    my $res = $self->_call('Network.disable');
 
-            return $cb ? $cb->($res) : $res;
-        }
-    );
+    $self->{network_enabled} = 0 if $res;
+
+    return $res;
 }
 
 # PAGE
-sub page_enable ( $self, $cb = undef ) {
-    return $cb ? $cb->( res 200 ) : res 200 if $self->{page_enabled};
+sub page_enable ( $self ) {
+    return res 200 if $self->{page_enabled};
 
-    return $self->_call(
-        'Page.enable',
-        undef,
-        sub ($res) {
-            $self->{page_enabled} = 1 if $res;
+    my $res = $self->_call('Page.enable');
 
-            return $cb ? $cb->($res) : $res;
-        }
-    );
+    $self->{page_enabled} = 1 if $res;
+
+    return $res;
 }
 
 sub page_disable ( $self, $cb = undef ) {
-    return $cb ? $cb->( res 200 ) : res 200 if !$self->{page_enabled};
+    return res 200 if !$self->{page_enabled};
 
-    return $self->_call(
-        'Page.disable',
-        undef,
-        sub ($res) {
-            $self->{page_enabled} = 0 if $res;
+    my $res = $self->_call('Page.disable');
 
-            return $cb ? $cb->($res) : $res;
-        }
-    );
+    $self->{page_enabled} = 0 if $res;
+
+    return $res;
 }
 
 # COOKIES
-sub get_cookies ( $self, $cb = undef ) {
-    weaken $self;
+sub get_cookies ( $self ) {
+    my $res = $self->_call('Network.getCookies');
 
-    return $self->_call(
-        'Network.getCookies',
-        undef,
-        sub ( $res ) {
-            my $cookies;
+    my $cookies;
 
-            if ( !$res ) {
-                warn $res;
-            }
-            else {
-                $cookies = defined $self ? $self->convert_cookies( $res->{data}->{cookies} ) : undef;
-            }
+    if ( !$res ) {
+        warn $res;
+    }
+    else {
+        $cookies = $self->convert_cookies( $res->{data}->{cookies} );
+    }
 
-            return $cb ? $cb->($cookies) : $cookies;
-        }
-    );
+    return $cookies;
 }
 
 sub convert_cookies ( $self, $chrome_cookies ) {
@@ -232,12 +209,8 @@ sub convert_cookies ( $self, $chrome_cookies ) {
 # format, png, jpeg
 # quality, 0 .. 100, for jpeg only
 # clip, { x, y, width, height, scale }
-sub get_screenshot ( $self, @args ) {
-    my $cb = is_plain_coderef $args[-1] ? pop @args : undef;
-
-    my $args = {@args};
-
-    if ( delete $args->{full} ) {
+sub get_screenshot ( $self, %args ) {
+    if ( delete $args{full} ) {
         my $metrics = $self->_call('Page.getLayoutMetrics');
 
         my $res = $self->_call(
@@ -248,24 +221,21 @@ sub get_screenshot ( $self, @args ) {
         );
     }
 
-    return $self->_call(
-        'Page.captureScreenshot',
-        $args,
-        sub ( $res ) {
-            my $img;
+    my $res = $self->_call( 'Page.captureScreenshot', \%args, );
 
-            if ($res) {
-                $img = from_b64 $res->{data}->{data};
-            }
-            else {
-                warn $res;
-            }
+    my $img;
 
-            return $cb ? $cb->($img) : $img;
-        }
-    );
+    if ($res) {
+        $img = from_b64 $res->{data}->{data};
+    }
+    else {
+        warn $res;
+    }
+
+    return $img;
 }
 
+# TODO !!! "Page.loadEventFired" is not fired
 sub navigate_to ( $self, $url, %args ) {
     my $listener     = $self->{listen}->{'Page.loadEventFired'};
     my $page_enabled = $self->{page_enabled};
@@ -280,19 +250,15 @@ sub navigate_to ( $self, $url, %args ) {
 
     my $cv = P->cv;
 
-    $self->{listen}->{'Page.loadEventFired'} = sub { $cv->() };
+    $self->listen( 'Page.loadEventFired', $cv );
 
-    $res = $self->_call( 'Page.navigate', { url => $url, %args }, undef );
+    $res = $self->_call( 'Page.navigate', { url => $url, %args } );
 
-    if ( !$res ) {
-        $self->page_disable if !$page_enabled;
-        $self->listen( 'Page.loadEventFired', $listener );
-
-        return $res;
-    }
+    goto CLEANUP if !$res;
 
     $cv->recv;
 
+  CLEANUP:
     $self->page_disable if !$page_enabled;
     $self->listen( 'Page.loadEventFired', $listener );
 
